from unittest import TestCase
from .context import *

from merklepy.path import build_path


class TestPath(TestCase):

    def test_build_path(self):
        expected = '0'
        found = build_path(index=1, size=2, depth=1)
        self.assertEqual(found, expected)

        expected = '10'
        found = build_path(index=1, size=4, depth=2)
        self.assertEqual(found, expected)

        expected = '110'
        found = build_path(index=1, size=5, depth=3)
        self.assertEqual(found, expected)

        expected = '101'
        found = build_path(index=2, size=5, depth=3)
        self.assertEqual(found, expected)

        expected = '1110'
        found = build_path(index=1, size=9, depth=4)
        self.assertEqual(found, expected)

        expected = '0111'
        found = build_path(index=8, size=9, depth=4)
        self.assertEqual(found, expected)



from unittest import TestCase
from .context import *
import base64

from merklepy.exception import InvalidMerkleProofError
from merklepy.hash import SHA_256
from merklepy.proof import Proof, proof_from


class TestProof(TestCase):

    def test_to_string(self):
        hashes = [
            bytes.fromhex(
                '1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef'),
            bytes.fromhex(
                'abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789')
        ]
        proof = Proof(hashes, '101', 5)
        expected = base64.b64encode(bytes(
            '1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdefabcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789.101.sha-256.5', 'utf-8')).decode('utf-8')
        found = proof.to_string()
        self.assertEqual(found, expected)

    def test_proof_from(self):
        proof = 'MTIzNDU2Nzg5MGFiY2RlZjEyMzQ1Njc4OTBhYmNkZWYxMjM0NTY3ODkwYWJjZGVmMTIzNDU2Nzg5MGFiY2RlZmFiY2RlZjAxMjM0NTY3ODlhYmNkZWYwMTIzNDU2Nzg5YWJjZGVmMDEyMzQ1Njc4OWFiY2RlZjAxMjM0NTY3ODkuMTAxLnNoYS0yNTYuNQ=='

        found = proof_from(proof)
        self.assertEqual(found.engine, SHA_256)
        self.assertEqual(found.path, '101')
        self.assertEqual(found.size, 5)
        self.assertEqual(len(found.trail), 2)
        self.assertEqual(found.trail[0], bytes.fromhex(
            '1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef'))
        self.assertEqual(found.trail[1], bytes.fromhex(
            'abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789'))
        self.assertEqual(found.to_string(), proof)

        self.assertRaises(InvalidMerkleProofError,
                          proof_from, 'not-a-valid-proof')



from unittest import TestCase
from .context import *

from merklepy.hash import SHA_256
from merklepy.options import MerkleTreeOptions


class TestOptions(TestCase):

    def test_merkle_tree_options(self):
        found = MerkleTreeOptions()
        self.assertFalse(found.doubleHash)
        self.assertEqual(found.engine, SHA_256)
        self.assertFalse(found.sort)

        defaultOptions = found.to_json()
        self.assertEqual(
            defaultOptions, '{"doubleHash":false,"engine":"sha-256","sort":false}')



from unittest import TestCase
from .context import *

from merklepy.exception import EmptyTreeError, InvalidJSONError, TreeNotBuildError
from merklepy.hash import SHA_256, build_hash_function
from merklepy.proof import proof_from
from merklepy.tree import MerkleTree, tree_from


class TestMerkleTree(TestCase):

    def test_add_leaves(self):
        tree = MerkleTree()
        self.assertRaises(EmptyTreeError, tree.add_leaves, True)
        self.assertRaises(EmptyTreeError, tree.add_leaves,
                          False, bytes('123', 'utf-8'))

    def test_depth(self):
        tree = MerkleTree()
        self.assertRaises(TreeNotBuildError, tree.depth)

        hashes = [
            bytes.fromhex(
                '1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef'),
            bytes.fromhex(
                'abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789')
        ]
        tree.add_leaves(False, *hashes)
        found = tree.depth()
        self.assertEqual(found, 1)

    def test_get_proof(self):
        data = list(map(lambda x: bytes(x, 'utf-8'), [
                    'data1', 'data2', 'data3', 'data4', 'data5']))
        tree = MerkleTree()
        maybe_proofs = tree.add_leaves(True, *data)

        proof1 = tree.get_proof(_sha256(bytes('data1', 'utf-8')))
        self.assertIsNotNone(proof1)
        self.assertEqual(proof1.to_string(), maybe_proofs[0].to_string())

        proof2 = tree.get_proof(_sha256(bytes('data2', 'utf-8')))
        self.assertIsNotNone(proof2)
        self.assertEqual(proof2.path, '110')
        self.assertEqual(len(proof2.trail), 3)
        self.assertEqual(proof2.to_string(), maybe_proofs[1].to_string())

        proof5 = tree.get_proof(_sha256(bytes('data5', 'utf-8')))
        self.assertIsNotNone(proof5)
        self.assertEqual(proof5.to_string(), maybe_proofs[4].to_string())

        tree2 = MerkleTree()
        tree2.add_leaves(True, *data)
        proof1_2 = tree2.get_proof(_sha256(bytes('data1', 'utf-8')))
        self.assertEqual(proof1_2.to_string(), proof1.to_string())

    def test_to_json(self):
        empty = MerkleTree()
        found = empty.to_json()
        self.assertEqual(
            found, '{"options":{"doubleHash":false,"engine":"sha-256","sort":false},"leaves":[]}')

        data = list(map(lambda x: bytes(x, 'utf-8'), [
                    'data1', 'data2', 'data3', 'data4', 'data5']))
        tree = MerkleTree()
        tree.add_leaves(True, *data)
        expected = '{"options":{"doubleHash":false,"engine":"sha-256","sort":false},"leaves":["5b41362bc82b7f3d56edc5a306db22105707d01ff4819e26faef9724a2d406c9","d98cf53e0c8b77c14a96358d5b69584225b4bb9026423cbc2f7b0161894c402c","f60f2d65da046fcaaf8a10bd96b5630104b629e111aff46ce89792e1caa11b18","02c6edc2ad3e1f2f9a9c8fea18c0702c4d2d753440315037bc7f84ea4bba2542","e195da4c40f26b85eb2b622e1c0d1ce73d4d8bf4183cd808d39a57e855093446"]}'
        found = tree.to_json()
        self.assertEqual(found, expected)

    def test_tree_from(self):
        empty = '{"options":{"doubleHash":false,"engine":"sha-256","sort":false},"leaves":[]}'
        self.assertRaises(InvalidJSONError, tree_from, empty)

        json = '{"options":{"doubleHash":false,"engine":"sha-256","sort":false},"leaves":["5b41362bc82b7f3d56edc5a306db22105707d01ff4819e26faef9724a2d406c9","d98cf53e0c8b77c14a96358d5b69584225b4bb9026423cbc2f7b0161894c402c","f60f2d65da046fcaaf8a10bd96b5630104b629e111aff46ce89792e1caa11b18","02c6edc2ad3e1f2f9a9c8fea18c0702c4d2d753440315037bc7f84ea4bba2542","e195da4c40f26b85eb2b622e1c0d1ce73d4d8bf4183cd808d39a57e855093446"]}'
        found = tree_from(json)
        self.assertEqual(found.get_root_hash(
        ), 'e9e1bc4a10c502ef995ede1914b0186ed288b8dde80c8c533a0f93a96490f995')
        proof = found.get_proof(bytes.fromhex(
            'd98cf53e0c8b77c14a96358d5b69584225b4bb9026423cbc2f7b0161894c402c'))
        self.assertIsNotNone(proof)

    def test_validate_proof(self):
        to_prove = proof_from('ZTE5NWRhNGM0MGYyNmI4NWViMmI2MjJlMWMwZDFjZTczZDRkOGJmNDE4M2NkODA4ZDM5YTU3ZTg1NTA5MzQ0NmFhNzVjZDVlMjUzMWYwNzJjNzAwN2JiMTkxZmViZGNkYmUyM2Q5YTRhZTMwY2RiYjg0Y2I1YTg2OWNlOWFiODM1YjQxMzYyYmM4MmI3ZjNkNTZlZGM1YTMwNmRiMjIxMDU3MDdkMDFmZjQ4MTllMjZmYWVmOTcyNGEyZDQwNmM5LjExMC5zaGEtMjU2LjU=')
        json = '{"options":{"doubleHash":false,"engine":"sha-256","sort":false},"leaves":["5b41362bc82b7f3d56edc5a306db22105707d01ff4819e26faef9724a2d406c9","d98cf53e0c8b77c14a96358d5b69584225b4bb9026423cbc2f7b0161894c402c","f60f2d65da046fcaaf8a10bd96b5630104b629e111aff46ce89792e1caa11b18","02c6edc2ad3e1f2f9a9c8fea18c0702c4d2d753440315037bc7f84ea4bba2542","e195da4c40f26b85eb2b622e1c0d1ce73d4d8bf4183cd808d39a57e855093446"]}'

        tree = tree_from(json)
        # It's supposed to be tested somehow, even though it's not part of the proof per se
        self.assertEqual(tree.size(), to_prove.size)
        found = tree.validate_proof(to_prove, _sha256(bytes(
            'data2', 'utf-8')), 'e9e1bc4a10c502ef995ede1914b0186ed288b8dde80c8c533a0f93a96490f995')
        self.assertTrue(found)

        # refuse an invalid data
        found = tree.validate_proof(to_prove, _sha256(bytes(
            'data6', 'utf-8')), 'e9e1bc4a10c502ef995ede1914b0186ed288b8dde80c8c533a0f93a96490f995')
        self.assertFalse(found)

        # refuse an invalid proof even for an existing data
        found = tree.validate_proof(to_prove, _sha256(bytes(
            'data1', 'utf-8')), 'e9e1bc4a10c502ef995ede1914b0186ed288b8dde80c8c533a0f93a96490f995')
        self.assertFalse(found)


_sha256 = build_hash_function(SHA_256)



from functools import reduce
import json
from typing import Optional

from merklepy.exception import EmptyTreeError, InvalidJSONError, TreeNotBuildError
from merklepy.hash import Hash, Hashes, build_hash_function, is_correct_hash, sort_hashes
from merklepy.options import MerkleTreeOptions
from merklepy.path import RIGHT, build_path
from merklepy.proof import Proof


class MerkleTree(object):

    def __init__(self, options: MerkleTreeOptions = MerkleTreeOptions()):
        self._isReady = False
        self._options = options
        self._leaves: Hashes = []
        self._leavesHex: list[str] = []
        self._levels: list[Hashes] = []
        self._hashFunction = build_hash_function(
            options.engine, options.doubleHash)

    def add_leaves(self, do_hash: bool, *args: bytes) -> list[Optional[Proof]]:
        """Add leaves, either sources (by passing `True` to the first parameter) or hashes"""
        self._isReady = False
        if len(args) == 0:
            raise EmptyTreeError()
        leaves = list(map(self._hashFunction, args)) if do_hash == True else list(
            filter(lambda h: is_correct_hash(h, self.get_engine()), args))
        if self._options.sort == True:
            self._sort()
        self._leaves = leaves
        self._leavesHex = list(map(lambda leaf: leaf.hex(), leaves))
        return self._make()

    def depth(self):
        """Returns the depth of the tree, ie. the number of levels excluding the root hash"""
        if self._isReady == False:
            raise TreeNotBuildError()
        return len(self._levels) - 1

    def get_engine(self):
        """Returns the name of the used hashing function"""
        return self._options.engine

    def get_proof(self, leaf: Hash) -> Optional[Proof]:
        """
        Try and retrieve the proof in the current Merkle tree for the passed hash

        Parameters
        ----------
        leaf : Hash
            The hashed data

        Returns
        -------
        Proof or None
            The correponding proof if any
        """
        if self._isReady == False:
            return None
        try:
            index = self._leaves.index(leaf)
            path = build_path(index, self.size(), self.depth())
            trail = list(
                map(lambda tuple: self._levels[tuple[0] + 1][tuple[1]], [(i, int(l)) for i, l in enumerate(path)]))
            if len(trail) == 0:
                return None
            return Proof(trail, path, self.size(), self.get_engine())
        except Exception as error:
            print(error)
            return None

    def get_root_hash(self):
        """Returns the hexadecimal representation of the root hash of the current Merkle tree"""
        if self._isReady == False:
            raise TreeNotBuildError()
        return self._levels[0][0].hex()

    def is_sorted(self):
        """Returns `true` if the current Merkle tree leaves are sorted, `false` otherwise"""
        return self._options.sort

    def size(self):
        """Returns the number of leaves"""
        return len(self._leaves)

    def to_json(self) -> str:
        """
        IMPORTANT: Use with caution!

        Returns
        -------
        str
            The JSON-stringified representation of the current Merkle tree

        Raises
        ------
        TreeNotBuiltError
            If there is no valid leaf added
        """
        if self._isReady == False and len(self._leaves) != 0:
            raise TreeNotBuildError()
        return f'{{"options":{self._options.to_json()},"leaves":{json.dumps(self._leavesHex, separators=(",", ":"))}}}'

    def use_double_hash(self):
        """Returns `true` if the current Merkle tree uses double hashing, `false` otherwise"""
        return self._options.doubleHash

    def validate_proof(self, proof: Proof, leaf: Hash, root_hash: str, rebuilding_proof: bool = False) -> bool:
        """
        Check that the passed proof matches the passed data using the passed root hash

        Parameters
        ----------
        proof : Proof
            The proof to use
        leaf : Hash
            The (hashed) data to check
        root_hash : str
            The hexadecimal representation of the root hash to compare to
        rebuilding_proof : bool, optional
            Set to `True` to use the method rebuilding the proof (default: `False`)

        Returns
        -------
        bool
            `True` if the proof is valid for the passed leaf
        """
        if self._isReady == False or root_hash != self.get_root_hash():
            return False
        if rebuilding_proof == True:
            rebuilt = self.get_proof(leaf)
            if rebuilt is None:
                return False
            return rebuilt.to_string() == proof.to_string()
        else:
            path = proof.path[::-1]
            trail = proof.trail[::-1]
            hFn = self._hashFunction
            proved = reduce(lambda h, tuple: hFn(tuple[1] + h) if path[tuple[0]] == RIGHT else hFn(
                h + tuple[1]), [(idx, current) for idx, current in enumerate(trail)], leaf)
            return proved.hex() == root_hash

    # For internal use only

    def _make(self):
        if len(self._leaves) == 0:
            raise EmptyTreeError()

        # Build the actual tree
        self._levels.insert(0, self._leaves)
        while len(self._levels[0]) > 1:
            self._levels.insert(0, self._next_level())
        if len(self._levels) == 0:
            return []
        self._isReady = True

        # Retrieve the proofs
        return list(map(self.get_proof, self._leaves))

    def _next_level(self):
        nodes = []
        from_level = self._levels[0]
        from_level_count = len(from_level)
        for i in range(0, from_level_count, 2):
            if i + 1 <= from_level_count - 1:
                nodes.append(self._hashFunction(
                    from_level[i] + from_level[i+1]))
            else:
                # Odd number promoted to the next level
                nodes.append(from_level[i])
        return nodes

    def _sort(self):
        self._leaves = sort_hashes(self._leaves)


def tree_from(json_string: str) -> MerkleTree:
    """Build a `MerkleTree` instance from the passed JSON string"""
    try:
        js = json.loads(json_string)
        options = MerkleTreeOptions() if 'options' not in js else MerkleTreeOptions(
            js['options']['doubleHash'], js['options']['engine'], js['options']['sort'])
        tree = MerkleTree(options)
        if 'leaves' not in js or len(js['leaves']) == 0:
            raise EmptyTreeError()
        leaves = list(map(bytes.fromhex, js['leaves']))
        tree.add_leaves(False, *leaves)
        return tree
    except Exception as error:
        raise InvalidJSONError(json_string) from error



import hashlib
import re
from typing import Callable

from merklepy.exception import InvalidEngineError

Hash = bytes
Hashes = list[Hash]

HashFunction = Callable[[bytes], Hash]


# Supported hash functions
SHA_256 = 'sha-256'

regex_sha256 = re.compile(r"""^[a-f0-9]{64}$""", re.I)


def build_hash_function(engine: str, doubleHash: bool = False) -> HashFunction:
    if engine == SHA_256:
        def Sha256(item: bytes) -> Hash:
            return hashlib.sha256(hashlib.sha256(item).digest()).digest() if doubleHash else hashlib.sha256(item).digest()
        return Sha256
    else:
        raise InvalidEngineError(engine)


def is_correct_hash(h: bytes, engine: str) -> bool:
    """Determine if it's a hash"""
    if engine == SHA_256:
        return regex_sha256.match(h.hex())
    else:
        return False


def sort_hashes(input: Hashes) -> Hashes:
    """Lexicographically sort the hexadecimal representation of the passed hashes"""
    return sorted(input)



from unittest import TestCase
from .context import *

from merklepy.exception import InvalidEngineError
from merklepy.hash import build_hash_function, is_correct_hash, SHA_256, sort_hashes


class TestHash(TestCase):

    def test_build_hash_function(self):
        sha256 = build_hash_function(SHA_256)
        expected = '9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08'
        found = sha256(bytes('test', 'utf-8'))
        self.assertEqual(found.hex(), expected)

        doubleSha256 = build_hash_function(SHA_256, True)
        expected = '954d5a49fd70d9b8bcdb35d252267829957f7ef7fa6c74f88419bdc5e82209f4'
        found = doubleSha256(bytes('test', 'utf-8'))
        self.assertEqual(found.hex(), expected)

        self.assertRaises(InvalidEngineError,
                          build_hash_function, 'wrong-engine')

    def test_is_correct_hash(self):
        correct = bytes.fromhex(
            '1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef')
        found = is_correct_hash(correct, SHA_256)
        self.assertTrue(found)

        incorrect = bytes('incorrect', 'utf-8')
        found = is_correct_hash(incorrect, SHA_256)
        self.assertFalse(found)

        found = is_correct_hash(correct, 'wrong-engine')
        self.assertFalse(found)

    def test_sort_hashes(self):
        hashes = [
            bytes.fromhex(
                'abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789'),
            bytes.fromhex(
                '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'),
            bytes.fromhex(
                'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        ]
        expected = [
            bytes.fromhex(
                '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'),
            bytes.fromhex(
                'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'),
            bytes.fromhex(
                'abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789')
        ]
        found = sort_hashes(hashes)
        self.assertEqual(len(found), 3)
        self.assertEqual(
            found[0].hex(),
            '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
        )
        self.assertEqual(found, expected)
